import unittest
import weaviate

class TestVersion(unittest.TestCase):

    def test_version(self):
        """
        Test the `__version__` global variable.
        """

        self.assertEqual(weaviate.__version__, "2.1.0rc0", "Check if the version is set correctly!")



import copy
from abc import ABC, abstractmethod
from typing import List, Sequence
from weaviate.util import get_valid_uuid, get_vector


class BatchRequest(ABC):
    """
    BatchRequest abstract class used as a interface for batch requests.
    """

    @abstractmethod
    def __len__(self):
        """
        This method should me implemented by all inheriting classes.
        """

    @abstractmethod
    def add(self, *args, **kwargs):
        """
        This method should me implemented by all inheriting classes.
        """

    @abstractmethod
    def get_request_body(self):
        """
        This method should me implemented by all inheriting classes.
        """


class ReferenceBatchRequest(BatchRequest):
    """
    Collect references to add them in one request to weaviate.
    Caution this request will miss some validations to be faster.
    """

    def __init__(self):
        self._from_object_class_names = []
        self._from_object_ids = []
        self._from_object_properties = []
        self._to_object_ids = []

    def __len__(self):
        return len(self._from_object_class_names)

    def add(self,
            from_object_uuid: str,
            from_object_class_name: str,
            from_property_name: str,
            to_object_uuid: str
        ) -> None:
        """
        Add one reference to this batch.

        Parameters
        ----------
        from_object_uuid : str
            The UUID or URL of the object that should reference another object.
        from_object_class_name : str
            The name of the class that should reference another object.
        from_property_name : str
            The name of the property that contains the reference.
        to_object_uuid : str
            The UUID or URL of the object that is actually referenced.

        Raises
        ------
        TypeError
            If arguments are not of type str.
        """

        if not isinstance(from_object_class_name, str) or not isinstance(from_object_uuid, str) or\
            not isinstance(from_property_name, str) or not isinstance(to_object_uuid, str):
            raise TypeError('All arguments must be of type string')

        from_object_uuid = get_valid_uuid(from_object_uuid)
        to_object_uuid = get_valid_uuid(to_object_uuid)

        self._from_object_class_names.append(from_object_class_name)
        self._from_object_ids.append(from_object_uuid)
        self._from_object_properties.append(from_property_name)
        self._to_object_ids.append(to_object_uuid)

    def get_request_body(self) -> List[dict]:
        """
        Get request body as a list of dictionaries, where each dictionary
        is a weaviate-schema reference.

        Returns
        -------
        list
            A list of references as dictionaries.
        """

        batch_body = []
        for i in range(len(self)):
            batch_body.append(
                {
                "from": "weaviate://localhost/" + self._from_object_class_names[i] + "/"
                    + self._from_object_ids[i] + "/" + self._from_object_properties[i],
                "to": "weaviate://localhost/" + self._to_object_ids[i]
                }
            )
        return batch_body


class ObjectsBatchRequest(BatchRequest):
    """
    Collect objects for one batch request to weaviate.
    Caution this batch will not be validated through weaviate.
    """

    def __init__(self):
        self._objects = []

    def __len__(self):
        return len(self._objects)

    def add(self,
            data_object: dict,
            class_name: str,
            uuid: str=None,
            vector: Sequence=None
        ) -> None:
        """
        Add one object to this batch.

        Parameters
        ----------
        data_object : dict
            Object to be added as a dict datatype.
        class_name : str
            The name of the class this object belongs to.
        uuid : str, optional
            UUID of the object as a string, by default None
        vector: Sequence, optional
            The embedding of the object that should be created. Used only class objects that do not
            have a vectorization module. Supported types are `list`, 'numpy.ndarray`,
            `torch.Tensor` and `tf.Tensor`,
            by default None.

        Raises
        ------
        TypeError
            If an argument passed is not of an appropriate type.
        ValueError
            If 'uuid' is not of a propper form.
        """

        if not isinstance(data_object, dict):
            raise TypeError("Object must be of type dict")
        if not isinstance(class_name, str):
            raise TypeError("Class name must be of type str")

        batch_item = {
            "class": class_name,
            "properties": copy.deepcopy(data_object)
        }
        if uuid is not None:
            batch_item["id"] = get_valid_uuid(uuid)

        if vector is not None:
            batch_item["vector"] = get_vector(vector)

        self._objects.append(batch_item)

    def get_request_body(self) -> dict:
        """
        Get the request body as it is needed for weaviate

        Returns
        -------
        dict
            The request body as a dict.
        """

        return {
            "fields": [
                "ALL"
            ],
            "objects": self._objects
        }



from .auth import AuthClientCredentials, AuthClientPassword
from .exceptions import *
from .util import generate_local_beacon
from .batch import ReferenceBatchRequest, ObjectsBatchRequest
from .classification import Classification
from .client import Client
from .version import __version__


"""
GraphQL filters for `Get` and `Aggregate` commands.
GraphQL abstract class for GraphQL commands to inherit from.
"""
import json
import sys
from copy import deepcopy
from typing import Optional
from abc import ABC, abstractmethod
from weaviate.connect import REST_METHOD_POST, Connection
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError

class GraphQL(ABC):
    """
    A base abstract class for GraphQL commands, such as Get, Aggregate.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a GraphQL abstract class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    @abstractmethod
    def build(self) -> str:
        """
        Build method to be overloaded by the child classes. It should return the
        GraphQL query as a str.

        Returns
        -------
        str
            The query.
        """

    def do(self) -> dict:
        """
        Builds and runs the query.

        Returns
        -------
        dict
            The response of the query.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        query = self.build()

        try:
            response = self._connection.run_rest("/graphql", REST_METHOD_POST, {"query": query})
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, query was not successful.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 200:
            return response.json()  # success
        raise UnexpectedStatusCodeException("Query was not successful", response)


class NearText:
    """
    NearText class used to filter weaviate objects.
    """

    def __init__(self, content: dict):
        """
        Initialize a NearText class instance.

        Parameters
        ----------
        content : dict
            The content of the `nearText` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content'  has key "certainty" but the value is not float.
        """

        if not isinstance(content, dict):
            raise TypeError(f"{self.__class__.__name__} filter is expected to "
                f"be type dict but was {type(content)}")
        _content = deepcopy(content)
        _check_concept(_content)
        self.concepts = _content["concepts"]
        self.certainty: Optional[float] = None
        self.move_to: Optional[dict] = None
        self.move_away_from: Optional[dict] = None

        if "certainty" in _content:
            if not isinstance(_content["certainty"], float):
                raise TypeError("certainty is expected to be a float but was "
                    f"{type(_content['certainty'])}")

            self.certainty = _content["certainty"]

        if "moveTo" in _content:
            _check_direction_clause(_content["moveTo"])
            self.move_to = _content["moveTo"]

        if "moveAwayFrom" in _content:
            _check_direction_clause(_content["moveAwayFrom"])
            self.move_away_from = _content["moveAwayFrom"]

    def __str__(self):
        near_text = f'nearText: {{concepts: {json.dumps(self.concepts)}'
        if self.certainty is not None:
            near_text += f' certainty: {str(self.certainty)}'
        if self.move_to is not None:
            near_text += (
                f' moveTo: {{concepts: {json.dumps(self.move_to["concepts"])} ' +\
                f'force: {self.move_to["force"]}}}'
            )
        if self.move_away_from is not None:
            near_text += (
                f' moveAwayFrom: {{concepts: {json.dumps(self.move_away_from["concepts"])} ' +\
                        f'force: {self.move_away_from["force"]}}}'
            )
        return near_text + '} '


class NearVector:
    """
    NearVector class used to filter weaviate objects.
    """

    def __init__(self, content: dict):
        """
        Initialize a NearVector class instance.

        Parameters
        ----------
        content : list
            The content of the `nearVector` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If 'content' does not contain "vector".
        TypeError
            If 'content["vector"]' is not of type list.
        AttributeError
            If invalid 'content' keys are provided.
        """

        if not isinstance(content, dict):
            raise TypeError(f"{self.__class__.__name__} filter is expected to "
                f"be type dict but was {type(content)}")

        _content = deepcopy(content)
        _check_vector(_content)
        self.vector = _content['vector']
        self.certainty: Optional[float] = None

        # Check optional fields

        if "certainty" in content:
            if not isinstance(_content["certainty"], float):
                raise TypeError("certainty is expected to be a float but was "
                            f"{type(_content['certainty'])}")
            self.certainty = _content["certainty"]

    def __str__(self):
        near_vector = f'nearVector: {{vector: {json.dumps(self.vector)}'
        if self.certainty is not None:
            near_vector += f' certainty: {self.certainty}'
        return near_vector + '} '


class WhereFilter:
    """
    WhereFilter class used to filter weaviate objects.
    """

    def __init__(self, content: dict):
        """
        Initialize a WhereFilter class instance.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If a mandatory key is missing in the filter content.
        """

        if not isinstance(content, dict):
            raise TypeError(f"{self.__class__.__name__} is expected to be type dict but "
                f"was {type(content)}")

        if "path" in content:
            self.is_filter = True
            self._parse_filter(content)
        elif "operands" in content:
            self.is_filter = False
            self._parse_operator(content)
        else:
            raise ValueError("Filter is missing required fileds `path` or `operands`."
                f" Given: {content}")

    def _parse_filter(self, content: dict) -> None:
        """
        Set filter fields for the WhereFilter.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        ValueError
            If 'content' is missing required fields.
        """

        if "operator" not in content:
            raise ValueError("Filter is missing required filed `operator`. "
                f"Given: {content}")

        self.path = json.dumps(content["path"])
        self.operator = content["operator"]
        self.value_type = _find_value_type(content)
        self.value = content[self.value_type]

    def _parse_operator(self, content: dict) -> None:
        """
        Set operator fields for the WhereFilter.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        ValueError
            If 'content' is missing required fields.
        """

        if "operator" not in content:
            raise ValueError("Filter is missing required filed `operator`."
                f" Given: {content}")
        _content = deepcopy(content)
        self.operator = _content["operator"]
        self.operands = []
        for operand in _content["operands"]:
            self.operands.append(WhereFilter(operand))

    def __str__(self):
        if self.is_filter:
            gql = f'where: {{path: {self.path} operator: {self.operator} {self.value_type}: '
            if self.value_type in ["valueInt", "valueNumber"]:
                gql += f'{self.value}}}'
            elif self.value_type == "valueBoolean":
                bool_value = str(self.value).lower()
                gql += f'{bool_value}}}'
            elif self.value_type == "valueGeoRange":
                geo_value = json.dumps(self.value)
                gql += f'{geo_value}}}'
            else:
                gql += f'"{self.value}"}}'
            return gql + ' '

        operands_str = []
        for operand in self.operands:
            # remove the `where: ` from the operands and the last space
            operands_str.append(str(operand)[7:-1])
        operands = ", ".join(operands_str)
        return f'where: {{operator: {self.operator} operands: [{operands}]}} '


def _check_direction_clause(direction: dict) -> dict:
    """
    Validate the direction sub clause.

    Parameters
    ----------
    direction : dict
        A sub clause of the Explore filter.

    Raises
    ------
    TypeError
        If 'direction' is not a dict.
    TypeError
        If the value of the "force" key is not float.
    ValueError
        If no "force" key in the 'direction'.
    """

    if not isinstance(direction, dict):
        raise TypeError(f"move clause should be dict but was {type(direction)}")
    _check_concept(direction)
    if not "force" in direction:
        raise ValueError("move clause needs to state a force")
    if not isinstance(direction["force"], float):
        raise TypeError(f"force should be float but was {type(direction['force'])}")


def _check_concept(content: dict) -> None:
    """
    Validate the concept sub clause.

    Parameters
    ----------
    content : dict
        An Explore (sub) clause to check for 'concepts'.

    Raises
    ------
    ValueError
        If no "concepts" key in the 'content' dict.
    TypeError
        If the value of the  "concepts" is of wrong type.
    """

    if "concepts" not in content:
        raise ValueError("No concepts in content")

    if not isinstance(content["concepts"], (list, str)):
        raise TypeError(f"Concepts must be of type list or str, not {type(content['concepts'])}")
    if isinstance(content["concepts"], str):
        content["concepts"] = [content["concepts"]]


def _check_vector(content: dict) -> None:
    """
    Validate the vector of the nearVector.

    Parameters
    ----------
    content : dict
        A nearVector clause to validate.

    Raises
    ------
    ValueError
        If no "vector" key in the 'content' dict.
    TypeError
        If the value of the  "vector" is of wrong type.
    """

    if "vector" not in content:
        raise ValueError("No 'vector' key in `content` argument.")
    if not isinstance(content["vector"], list):
        raise TypeError(f"'vector' key is expected to be type `list` but was {type(content['vector'])}")


def _find_value_type(content: dict) -> str:
    """
    Find the correct type of the content.

    Parameters
    ----------
    content : dict
        The content for which to find the appropriate data type.

    Returns
    -------
    str
        The correct data type.

    Raises
    ------
    ValueError
        If missing required fields.
    """

    if "valueString" in content:
        to_return = "valueString"
    elif "valueText" in content:
        to_return = "valueText"
    elif "valueInt" in content:
        to_return = "valueInt"
    elif "valueNumber" in content:
        to_return = "valueNumber"
    elif "valueDate" in content:
        to_return = "valueDate"
    elif "valueBoolean" in content:
        to_return = "valueBoolean"
    elif "valueGeoRange" in content:
        to_return = "valueGeoRange"
    else:
        raise ValueError(f"Filter is missing required fileds: {content}")
    return to_return



import time
from typing import Callable
from requests.exceptions import ReadTimeout, Timeout
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from weaviate.batch.requests import BatchRequest


class SubmitBatchesException(Exception):
    """
    Submit Batch Exception.
    """


class SubmitBatches:
    """
    SubmitBatcher class used to submit data in batcher from the Batcher object.
    """

    def __init__(self,
            max_backoff_time: int = 300,
            max_request_retries: int = 4,
            verbose_enabled: bool = False
        ):
        """
        Initialize a SubmitBatches class instance.

        Parameters
        ----------
        max_backoff_time : int, optional
            Max time used in the exponential backoff, by default 300.
        max_request_retries : int, optional
            States how often a request is retried before it counts as failed, by default 4.
        verbose_enabled : bool, optional
            If True errors will be printed directly to stdout, by default False.
        """

        self._backoff_time = 10
        self._backoff_count = 0
        self._max_backoff_time = max_backoff_time
        self._result_collection = []
        self._verbose_enabled = verbose_enabled
        self._max_request_retries = max_request_retries

    def pop_results(self) -> list:
        """
        Pops the results of the submissions.

        Returns
        -------
        list
            The results from all submitted updates since the last pop.
        """

        return_value = self._result_collection
        self._result_collection = []
        return return_value

    def submit_update(self, create_func: Callable[[BatchRequest], None], data: BatchRequest):
        """
        Parameters
        ----------
        create_func: Callable[[BatchRequest], None]
            The function that should be called with the data.
        data: BatchRequest
            The Batch object that should be used as parameter to the `create_func`.

        Raises
        ------
        SubmitBatchesException
            If the batch could not be submitted.
        """

        if len(data) == 0:
            return

        retry_counter = 0
        while not self._submit_data(create_func, data):
            self._sleep_backoff()
            retry_counter += 1
            if retry_counter >= self._max_request_retries:
                raise SubmitBatchesException("Could not submit data")

        self._reset_backoff()

    def _submit_data(self, create_func: Callable[[BatchRequest], None], data: BatchRequest):
        """
        Submits `data` using the `create_func` and appends the result to the result collection.

        Parameters
        ----------
        create_func: Callable[[BatchRequest], None]
            The function that should be called with the data.
        data: BatchRequest
            The Batch object that should be used as parameter to the `create_func`.

        Return
        bool
            False if failed, True if successful.
        """

        try:
            result = create_func(data)
        except (RequestsConnectionError, Timeout, ReadTimeout,\
                        UnexpectedStatusCodeException) as error:
            print("Exception in creating data: ", error)
            return False
        else:
            if isinstance(result, list):
                self._result_collection += result
                self._print_errors(result)
            return True

    def _sleep_backoff(self):
        """
        Calculates an exponential backoff and sleeps for the calculated time.
        Is limited by the max_backoff_time settable in the constructor
        """

        self._backoff_time = self._backoff_time + self._backoff_count * self._backoff_time
        if self._backoff_time > self._max_backoff_time:
            self._backoff_time = self._max_backoff_time
        else:
            self._backoff_count += 1
        time.sleep(self._backoff_time)

    def _reset_backoff(self):
        """
        Reset backoff values.
        """

        self._backoff_time = 10
        self._backoff_count = 0

    def _print_errors(self, request_result: dict):
        """
        Print request results, only for verbose option of the Batcher class.

        Parameters
        ----------
        request_result : dict
            Request result of the batch submission.
        """

        if not self._verbose_enabled:
            return

        for item in request_result:
            if len(item['result']) > 0:
                if not 'status' in item['result']:
                    print(item)
                    return

                if item['result']['status'] != 'SUCCESS':
                    print(item)



import unittest
from unittest.mock import patch, Mock
from weaviate.util import * 
from weaviate.util import  _get_dict_from_object, _is_sub_schema, _get_valid_timeout_config
from test.util import check_startswith_error_message, check_error_message


schema_set = {
    "classes": [
        {
            "class": "Ollie",
            "properties": [{"name": "height"}]
        },
        {
            "class": "Shuvit",
            "properties": [{"name": "direction"}]
        },
        {
            "class": "Board",
            "properties": [{"name": "brand"},
                            {"name": "art"},
                            {"name": "size"}]
        },
        {
            "class": "Truck",
            "properties": [{"name": "name"},
                            {"name": "height"}]
        }
    ],
}

schema_set_extended_prop = {
    "classes": [
        {
            "class": "Ollie",
            "properties": [{"name": "height", "name": "weight"}]
        },
        {
            "class": "Shuvit",
            "properties": [{"name": "direction"}]
        },
        {
            "class": "Board",
            "properties": [{"name": "brand"},
                            {"name": "art"},
                            {"name": "size"}]
        },
        {
            "class": "Truck",
            "properties": [{"name": "name"},
                            {"name": "height"}]
        }
    ],
}

schema_sub_set = {
    "classes": [
        {
            "class": "Ollie",
            "properties": [{"name": "height"}]
        },
        {
            "class": "Board",
            "properties": [{"name": "brand"},
                            {"name": "art"},
                            {"name": "size"}]
        }
    ],
}

disjoint_set = {
    "classes": [
        {
            "class": "Manual",
            "properties": [{"name": "nose"}]
        },
        {
            "class": "Bearings",
            "properties": [{"name": "brand"}]
        }
    ],
}

partial_set = {
    "classes": [
        {
            "class": "Board",
            "properties": [{"name": "brand"},
                            {"name": "art"},
                            {"name": "size"}]
        },
        {
            "class": "Truck",
            "properties": [{"name": "name"},
                            {"name": "height"}]
        },
        {
            "class": "Bearings",
            "properties": [{"name": "brand"}]
        },
        {
            "class": "Ollie",
            "properties": [{"name": "height"}]
        },
        {
            "class": "Shuvit",
            "properties": [{"name": "direction"}]
        },
        {
            "class": "Manual",
            "properties": [{"name": "nose"}]
        }
    ],
}

schema_company = {
  "classes": [
    {
      "class": "Company",
      "description": "A business that acts in the market",
      "properties": [
        {
          "name": "name",
          "description": "The name under which the company is known",
          "dataType": ["text"]
        },
        {
          "name": "legalBody",
          "description": "The legal body under which the company maintains its business",
          "dataType": ["text"]
        },
        {
          "name": "hasEmployee",
          "description": "The employees of the company",
          "dataType": ["Employee"]
        }
      ]
    },
    {
      "class": "Employee",
      "description": "An employee of the company",
      "properties": [
        {
          "name": "name",
          "description": "The name of the employee",
          "dataType": ["text"]
        },
        {
          "name": "job",
          "description": "the job description of the employee",
          "dataType": ["text"]
        },
        {
          "name": "yearsInTheCompany",
          "description": "The number of years this employee has worked in the company",
          "dataType": ["int"]
        }
      ]
    }
  ]
}


class TestUtil(unittest.TestCase):

    def test_generate_local_beacon(self):
        """
        Test the `generate_local_beacon` function.
        """

        type_error_message = "Expected to_object_uuid of type str"
        value_error_message = "Uuid does not have the propper form"
        # wrong data type
        with self.assertRaises(TypeError) as error:
            generate_local_beacon(None)
        check_error_message(self, error, type_error_message)
        # wrong value
        with self.assertRaises(ValueError) as error:
            generate_local_beacon("Leeroy Jenkins")
        check_error_message(self, error, value_error_message)

        beacon = generate_local_beacon("fcf33178-1b5d-5174-b2e7-04a2129dd35a")
        self.assertTrue("beacon" in beacon)
        self.assertEqual(beacon["beacon"], "weaviate://localhost/fcf33178-1b5d-5174-b2e7-04a2129dd35a")

        beacon = generate_local_beacon("fcf33178-1b5d-5174-b2e7-04a2129dd35b")
        self.assertTrue("beacon" in beacon)
        self.assertEqual(beacon["beacon"], "weaviate://localhost/fcf33178-1b5d-5174-b2e7-04a2129dd35b")

    def test__get_dict_from_object(self):
        """
        Test the `_get_dict_from_object` function.
        """

        none_error_message = "argument is None"
        file_error_message = "No file found at location "
        url_error_message = "Could not download file "
        type_error_message = ("Argument is not of the supported types. Supported types are "
                    "url or file path as string or schema as dict.")
        # test wrong type None
        with self.assertRaises(TypeError) as error:
            _get_dict_from_object(None)
        check_error_message(self, error, none_error_message)
        # wrong data type
        with self.assertRaises(TypeError) as error:
            _get_dict_from_object([{"key": 1234}])
        check_error_message(self, error, type_error_message)
        # wrong path
        with self.assertRaises(ValueError) as error:
            _get_dict_from_object("not_a_path_or_url.txt")
        check_error_message(self, error, file_error_message + "not_a_path_or_url.txt")
        # wrong URL or non existing one or failure of requests.get
        with patch('weaviate.util.requests') as mock_obj:
            result_mock = Mock()
            result_mock.status_code = 404
            mock_obj.get.return_value = result_mock
            with self.assertRaises(ValueError) as error:
                _get_dict_from_object("http://www.url.com")
            check_error_message(self, error, url_error_message + "http://www.url.com")
            mock_obj.get.assert_called()

        # valid calls
        self.assertEqual(_get_dict_from_object({"key": "val"}), {"key": "val"})
        # read from file
        path = '/'.join(__file__.split('/')[:-1])
        self.assertEqual(_get_dict_from_object(f'{path}/schema/schema_company.json'), schema_company)
        # read from URL
        path = "https://raw.githubusercontent.com/semi-technologies/weaviate-python-client/weaviate_v1/test/schema/schema_company.json"
        self.assertEqual(_get_dict_from_object(path), schema_company)

    def test_is_weaviate_object_url(self):
        """
        Test the `is_weaviate_object_url` function.
        """

        # valid formats
        self.assertTrue(
            is_weaviate_object_url("weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2"))
        self.assertTrue(
            is_weaviate_object_url("weaviate://some-domain.com/28f3f61b-b524-45e0-9bbe-2c1550bf73d2"))

        # invalid formats
        ## wrong argument data type
        self.assertFalse(
            is_weaviate_object_url(["weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2"]))
        ## wrong prefix, i.e. does not start with 'weaviate://' 
        self.assertFalse(
            is_weaviate_object_url("http://some-domain.com/28f3f61b-b524-45e0-9bbe-2c1550bf73d2"))
        ## wrong path, additional '/thing'
        self.assertFalse(
            is_weaviate_object_url("weaviate://localhost/things/f61b-b524-45e0-9bbe-2c1550bf73d2"))
        ## worng domain format
        self.assertFalse(
            is_weaviate_object_url("weaviate://some-INVALID-domain/28f3f61b-b524-45e0-9bbe-2c1550bf73d2"))
        # wrong UUID format
        self.assertFalse(
            is_weaviate_object_url("weaviate://localhost/UUID"))

    def test_is_object_url(self):
        """
        Test the `is_object_url` function.
        """

        # valid formats
        self.assertTrue(
            is_object_url("http://localhost:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446"))
        self.assertTrue(
            is_object_url("http://ramalamadingdong/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446"))

        # invalid formats
        ## wrong path, should be at least 3 subpaths to the object UUID
        self.assertFalse(
            is_object_url("objects/1c9cd584-88fe-5010-83d0-017cb3fcb446"))
        ## wrong '/v2', shoudl be '/v1'  
        self.assertFalse(
            is_object_url("http://localhost:8080/v2/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446"))
        ## wrong UUID format
        self.assertFalse(
            is_object_url("http://ramalamadingdong/v1/objects/1c9cd584-88fe-5010-83d0"))
        ## wrong objects path, instead of '/passions' should have been '/objects/
        self.assertFalse(
            is_object_url("http://localhost:8080/v1/passions/1c9cd584-88fe-5010-83d0-017cb3fcb446"))

    def test_get_valid_uuid(self):
        """
        Test the `get_valid_uuid` function.
        """

        # valid calls
        result = get_valid_uuid("weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2")
        self.assertEqual(result, "28f3f61b-b524-45e0-9bbe-2c1550bf73d2")

        result = get_valid_uuid("weaviate://otherhost.com/28f3f61b-b524-45e0-9bbe-2c1550bf73d2")
        self.assertEqual(result, "28f3f61b-b524-45e0-9bbe-2c1550bf73d2")

        result = get_valid_uuid("http://localhost:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446")
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        result = get_valid_uuid("http://otherhost_2:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446")
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        result = get_valid_uuid("1c9cd584-88fe-5010-83d0-017cb3fcb446")
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        # invalid formats
        type_error_message = "'uuid' must be of type str but was: "
        value_error_message = "Not valid 'uuid' or 'uuid' can not be extracted from value"
        ## neither an object URL nor a weaviate object URL
        with self.assertRaises(ValueError) as error:
            get_valid_uuid("http://localhost:8080/v1/1c9cd584-88fe-5010-83d0-017cb3fcb")
        check_error_message(self, error, value_error_message)

        # wrong UUID format
        with self.assertRaises(ValueError) as error:
            get_valid_uuid("http://localhost:8080/v1/objects/some-UUID")
        check_error_message(self, error, value_error_message)

        ## wrong '/v2', shoudl be '/v1'
        with self.assertRaises(ValueError) as error:
            get_valid_uuid("http://localhost:8080/v2/objects/1c9cd584-88fe-5010-83d0-017cb3fcb")
        check_error_message(self, error, value_error_message)

        ## wrong URL
        with self.assertRaises(ValueError) as error:
            get_valid_uuid("weaviate://INVALID_URL//1c9cd584-88fe-5010-83d0-017cb3fcb")
        check_error_message(self, error, value_error_message)

        ## wrong UUID data type
        with self.assertRaises(TypeError) as error:
            get_valid_uuid(12)
        check_error_message(self, error, type_error_message + str(int))

    def test_get_vector(self):
        """
        Test the `get_vector` function.
        """

        vector_list  = [1., 2., 3.]
        # vector is a list
        self.assertEqual(get_vector(vector_list), vector_list)

        # vector is a `torch.Tensor` or `numpy.ndarray`
        vector_mock = Mock()
        mock_tolist = Mock()
        mock_tolist.tolist.return_value = vector_list
        mock_squeeze = Mock(return_value = mock_tolist)
        vector_mock.squeeze = mock_squeeze
        self.assertEqual(get_vector(vector_mock), vector_list)

        # vector is a `tf.Tensor`
        vector_mock = Mock()
        mock_tolist = Mock()
        mock_squeeze = Mock()
        mock_tolist.tolist.return_value = vector_list
        mock_squeeze.squeeze.return_value = mock_tolist
        mock_numpy = Mock(return_value = mock_squeeze)
        vector_mock.numpy = mock_numpy
        vector_mock.squeeze = Mock(side_effect = AttributeError("TEST TensorFlow Tensor"))
        self.assertEqual(get_vector(vector_mock), vector_list)

        # invalid call
        type_error_message = ("The type of the 'vector' argument is not supported!\n"
                "Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` "
                "and `tf.Tensor`")
        with self.assertRaises(TypeError) as error:
            get_vector('[1., 2., 3.]')
        check_error_message(self, error, type_error_message)

    def test_get_domain_from_weaviate_url(self):
        """
        Test the `get_domain_from_weaviate_url` function.
        """

        uuid = "28f3f61b-b524-45e0-9bbe-2c1550bf73d2"
        self.assertEqual(get_domain_from_weaviate_url(f"weaviate://localhost/{uuid}"), "localhost")
        self.assertEqual(get_domain_from_weaviate_url(f"weaviate://otherhost/{uuid}"), "otherhost")

    def test__is_sub_schema(self):
        """
        Test the `_is_sub_schema` function.
        """

        self.assertTrue(_is_sub_schema(schema_set, schema_set))
        self.assertTrue(_is_sub_schema(schema_sub_set, schema_set))
        self.assertTrue(_is_sub_schema({}, schema_set))

        self.assertFalse(_is_sub_schema(disjoint_set, schema_set))
        self.assertFalse(_is_sub_schema(partial_set, schema_set))
        self.assertFalse(_is_sub_schema(schema_set_extended_prop, schema_set))

    def test__get_valid_timeout_config(self):
        """
        Test the `_get_valid_timeout_config` function.
        """

        # incalid calls 
        type_error_message = "'timeout_config' should be either a tuple or a list!"
        value_error_message = "'timeout_config' must be of length 2!"
        value_types_error_message = "'timeout_config' must be tupel of int"
        ## wrong type, None
        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config(None)
        check_error_message(self, error, type_error_message)

        ## wrong type, not list or tuple
        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config("(2, 13)")
        check_error_message(self, error, type_error_message)

        ## worng tuple length 3
        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config((1,2,3))
        check_error_message(self, error, value_error_message)

        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config([1, 2, 3])
        check_error_message(self, error, value_error_message)

        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config(tuple([1]))
        check_error_message(self, error, value_error_message)

        with self.assertRaises(ValueError) as error:
            _get_valid_timeout_config([1])
        check_error_message(self, error, value_error_message)

        ## wrong value types
        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config([1, 10.123])
        check_error_message(self, error, value_types_error_message)

        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config(["1", 10])
        check_error_message(self, error, value_types_error_message)

        with self.assertRaises(TypeError) as error:
            _get_valid_timeout_config(["1", "10"])
        check_error_message(self, error, value_types_error_message)

        # valid calls
        _get_valid_timeout_config((2, 20))
        _get_valid_timeout_config([20, 10])



"""
Triple schema as constant.
"""

triple_schema = {
  "classes": [
    {
      "class": "Subject",
      "description": "Subject",
      "properties": [
        {
          "name": "value",
          "description": "value of subject",
          "dataType": [
            "text"
          ],
        },
        {
          "name": "valueKey",
          "description": "valueKey of subject",
          "dataType": [
            "string"
          ],
        }
      ]
    },
    {
      "class": "Predicate",
      "description": "predicate",
      "properties": [
        {
          "name": "value",
          "description": "value",
          "dataType": [
            "text"
          ],
        },
        {
          "name": "valueKey",
          "description": "valueKey",
          "dataType": [
            "string"
          ],
        }
      ]
    },
    {
      "class": "Object",
      "description": "object",
      "properties": [
        {
          "name": "value",
          "description": "value",
          "dataType": [
            "text"
          ],
        },
        {
          "name": "valueKey",
          "description": "valueKey of subject",
          "dataType": [
            "string"
          ],
        }
      ]
    },
    {
      "class": "Triple",
      "description": "rdf triple",
      "properties": [
        {
          "name": "subject",
          "description": "subject of the triple",
          "dataType": [
            "Subject"
          ],
        },
        {
          "name": "predicate",
          "description": "predicate of the triple",
          "dataType": [
            "Predicate"
          ],
        },
        {
          "name": "object",
          "description": "object of the triple",
          "dataType": [
            "Object"
          ],
        }
      ]
    }
  ]
}



import unittest
import weaviate


class TestAuthentication(unittest.TestCase):

    def test_client_credentials(self):
        """
        Test AuthClientCredentials.
        """

        token = "testtoken4711"
        credentials = weaviate.AuthClientCredentials(token)
        request_body = credentials.get_credentials()
        self.assertTrue("grant_type" in request_body)
        self.assertTrue("client_secret" in request_body)

        self.assertEqual(request_body["client_secret"], token)
        self.assertEqual(request_body["grant_type"], "client_credentials")

    def test_user_password(self):
        """
        Test AuthClientPassword.
        """

        user = "@greenstalone"
        password = "testtoken4711"
        credentials = weaviate.AuthClientPassword(user, password)
        request_body = credentials.get_credentials()

        self.assertEqual(request_body["username"], user)
        self.assertEqual(request_body["password"], password)
        self.assertEqual(request_body["grant_type"], "password")



import unittest
from unittest.mock import patch, Mock
import weaviate
from weaviate import Client
from weaviate.exceptions import RequestsConnectionError
from test.util import mock_run_rest, replace_connection, check_error_message
        

class TestWeaviateClient(unittest.TestCase):

    def test___init__(self):
        """
        Test the `__init__` method.
        """

        Client._connection = Mock()
        type_error_message = "URL is expected to be string but is "
        # test incalid calls
        with self.assertRaises(TypeError) as error:
            Client(None)
        check_error_message(self, error, type_error_message + str(type(None)))
        with self.assertRaises(TypeError) as error:
            Client(42)
        check_error_message(self, error, type_error_message + str(int))

        # test valid calls
        with patch('weaviate.client.Connection') as mock_obj:
            Client("some_URL", auth_client_secret=None, timeout_config=(1, 2))
            mock_obj.assert_called_with(
                url='some_URL',
                auth_client_secret=None,
                timeout_config=(1,2)
                )
        with patch('weaviate.client.Connection') as mock_obj:
            Client("some_URL/", auth_client_secret=None, timeout_config=(5, 20))
            mock_obj.assert_called_with(
                url='some_URL',
                auth_client_secret=None,
                timeout_config=(5,20)
                )


    def test_is_ready(self):
        """
        Test the `is_ready` method.
        """

        client = Client("http://localhost:8080")
        # Request to weaviate returns 200
        connection_mock = mock_run_rest()
        client._connection = connection_mock
        self.assertTrue(client.is_ready())  # Should be true
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/.well-known/ready", weaviate.connect.REST_METHOD_GET)

        # Request to weaviate returns 404
        connection_mock = mock_run_rest(status_code=404)
        client._connection = connection_mock
        self.assertFalse(client.is_ready())  # Should be false
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/.well-known/ready", weaviate.connect.REST_METHOD_GET)

        # Test exception in connect
        connection_mock = mock_run_rest(side_effect=RequestsConnectionError("Test"))
        client._connection = connection_mock
        self.assertFalse(client.is_ready())
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/.well-known/ready", weaviate.connect.REST_METHOD_GET)

    def test_is_live(self):
        """
        Test the `is_live` method.
        """

        client = Client("http://localhost:8080")
        # Request to weaviate returns 200
        connection_mock = mock_run_rest()
        client._connection = connection_mock
        self.assertTrue(client.is_live())  # Should be true
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/.well-known/live", weaviate.connect.REST_METHOD_GET)

        # Request to weaviate returns 404
        connection_mock = mock_run_rest(status_code=404)
        client._connection = connection_mock
        self.assertFalse(client.is_live())  # Should be false
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/.well-known/live", weaviate.connect.REST_METHOD_GET)

    def test_get_meta(self):
        """
        Test the `get_meta` method.
        """

        client = Client("http://localhost:8080")
        # Request to weaviate returns 200
        connection_mock = mock_run_rest(return_json="OK!")
        client._connection = connection_mock
        self.assertEqual(client.get_meta(), "OK!")
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/meta", weaviate.connect.REST_METHOD_GET)

        # Request to weaviate returns 404
        connection_mock = mock_run_rest(status_code=404)
        client._connection = connection_mock
        with self.assertRaises(weaviate.UnexpectedStatusCodeException) as error:
            client.get_meta()
        error_message = "Meta endpoint! Unexpected status code: 404, with response body: None"
        check_error_message(self, error, error_message)
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/meta", weaviate.connect.REST_METHOD_GET)
    
    def test_get_open_id_configuration(self):
        """
        Test the `get_open_id_configuration` method.
        """

        client = Client("http://localhost:8080")
        # Request to weaviate returns 200
        connection_mock = mock_run_rest(return_json="OK!")
        client._connection = connection_mock
        self.assertEqual(client.get_open_id_configuration(), "OK!")
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/.well-known/openid-configuration", weaviate.connect.REST_METHOD_GET)

        
        # Request to weaviate returns 404
        connection_mock = mock_run_rest(status_code=404)
        client._connection = connection_mock
        self.assertIsNone(client.get_open_id_configuration())
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/.well-known/openid-configuration", weaviate.connect.REST_METHOD_GET)

        # Request to weaviate returns 204
        connection_mock = mock_run_rest(status_code=204)
        client._connection = connection_mock
        with self.assertRaises(weaviate.UnexpectedStatusCodeException) as error:
            client.get_open_id_configuration()
        error_message = f"Meta endpoint! Unexpected status code: 204, with response body: None"
        check_error_message(self, error, error_message)
        connection_mock.run_rest.assert_called()
        connection_mock.run_rest.assert_called_with("/.well-known/openid-configuration", weaviate.connect.REST_METHOD_GET)

    def test_timeout_config(self):
        """
        Test the `set_timeout_config` method.
        """

        client = Client("some_URL", auth_client_secret=None, timeout_config=(1, 2))
        self.assertEqual(client.timeout_config, (1, 2))
        client.timeout_config = (4, 20) #;)
        self.assertEqual(client.timeout_config, (4, 20))



from typing import Optional, Tuple
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from .connect import Connection, REST_METHOD_GET
from .classification import Classification
from .schema import Schema
from .contextionary import Contextionary
from .batch import Batch
from .data import DataObject
from .gql import Query
from .auth import AuthCredentials


class Client:
    """
    A python native weaviate client.
    """

    def __init__(self,
            url: str,
            auth_client_secret: AuthCredentials=None,
            timeout_config: Optional[Tuple[int, int]]=None
        ):
        """
        Initialize a Client class instance.

        Parameters
        ----------
        url : str
            The URL to the weaviate instance.
        auth_client_secret : weaviate.AuthCredentials, optional
            Authentification client secret, by default None.
        timeout_config : tuple(int, int), optional
            Set the timeout config as a tuple of (retries, time out seconds),
            by default None.

        Raises
        ------
        TypeError
            If arguments are of a wrong data type.
        """

        if not isinstance(url, str):
            raise TypeError("URL is expected to be string but is " + str(type(url)))
        if url.endswith("/"):
            # remove trailing slash
            url = url[:-1]

        self._connection = Connection(
            url=url,
            auth_client_secret=auth_client_secret,
            timeout_config=timeout_config
        )
        self.classification = Classification(self._connection)
        self.schema = Schema(self._connection)
        self.contextionary = Contextionary(self._connection)
        self.batch = Batch(self._connection)
        self.data_object = DataObject(self._connection)
        self.query = Query(self._connection)

    def is_ready(self) -> bool:
        """
        Ping weaviates ready state

        Returns
        -------
        bool
            True if weaviate is ready to accept requests,
            False otherwise.
        """

        try:
            response = self._connection.run_rest("/.well-known/ready", REST_METHOD_GET)
            if response.status_code == 200:
                return True
            return False
        except RequestsConnectionError:
            return False

    def is_live(self) -> bool:
        """
        Ping weaviates live state.

        Returns
        --------
        bool
            True if weaviate is live and should not be killed,
            False otherwise.
        """

        response = self._connection.run_rest("/.well-known/live", REST_METHOD_GET)
        if response.status_code == 200:
            return True
        return False

    def get_meta(self) -> dict:
        """
        Get the meta endpoint description of weaviate.

        Returns
        -------
        dict
            The dict describing the weaviate configuration.

        Raises
        ------
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        response = self._connection.run_rest("/meta", REST_METHOD_GET)
        if response.status_code == 200:
            return response.json()
        raise UnexpectedStatusCodeException("Meta endpoint", response)

    def get_open_id_configuration(self) -> Optional[dict]:
        """
        Get the openid-configuration.

        Returns
        -------
        dict
            The configuration or None if not configured.

        Raises
        ------
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        response = self._connection.run_rest("/.well-known/openid-configuration", REST_METHOD_GET)
        if response.status_code == 200:
            return response.json()
        if response.status_code == 404:
            return None
        raise UnexpectedStatusCodeException("Meta endpoint", response)

    @property
    def timeout_config(self):
        """
        Getter for `timeout_config`.
        """

        return self._connection.timeout_config

    @timeout_config.setter
    def timeout_config(self, timeout_config: Optional[Tuple[int, int]]):
        """
        Setter for `timeout_config`.

        Parameters
        ----------
        timeout_config : tuple(int, int) or list[int, int]
            Timeout config as a tuple of (retries, time out seconds).
        """

        self._connection.timeout_config = timeout_config



import unittest
from unittest.mock import Mock
from weaviate.exceptions import *


class TestExceptions(unittest.TestCase):

    def test_unexpected_status_code(self):
        """
        Test the `UnexpectedStatusCodeException` exception.
        """

        # with .json() exception raised
        response = Mock()
        response.json = Mock()
        response.json.side_effect = Exception("Test")
        response.status_code = 1234
        exception = UnexpectedStatusCodeException(message="Test message", response=response)

        self.assertEqual(exception.status_code, 1234)
        self.assertIsNone(exception.json)
        self.assertEqual(str(exception), "Test message! Unexpected status code: 1234, with response body: None")

        # with .json() value
        response = Mock()
        response.json = Mock()
        response.json.return_value = {"test" : "OK!"}
        response.status_code = 4321
        exception = UnexpectedStatusCodeException(message="Second test message", response=response)

        self.assertEqual(exception.status_code, 4321)
        self.assertEqual(exception.json, {"test" : "OK!"})
        self.assertEqual(str(exception), "Second test message! Unexpected status code: 4321, with response body: {'test': 'OK!'}")

    def test_object_already_exists(self):
        """
        Test the `ObjectAlreadyExistsException` exception.
        """

        exception = ObjectAlreadyExistsException("Test")
        self.assertEqual(str(exception), "Test")

    def test_authentication_failed(self):
        """
        Test the `AuthenticationFailedException` exception.
        """

        exception = AuthenticationFailedException("Test")
        self.assertEqual(str(exception), "Test")

    def test_schema_validation(self):
        """
        Test the `SchemaValidationException` exception.
        """

        exception = SchemaValidationException("Test")
        self.assertEqual(str(exception), "Test")